from typing import Any as _Any, override as _override

from leads.constant import SystemLiteral
from leads.context import Context
from leads.event import InterventionEvent, InterventionExitEvent
from leads.plugin.plugin import Plugin
from leads.types import OptionalSpeed as _OptionalSpeed


def dtcs_srw(context: Context,
             front_wheel_speed: _OptionalSpeed,
             rear_wheel_speed: _OptionalSpeed) -> InterventionEvent:
    if front_wheel_speed < rear_wheel_speed:
        context.overwrite_throttle(0)
        return InterventionEvent(context, SystemLiteral.DTCS, front_wheel_speed, rear_wheel_speed)
    return InterventionExitEvent(context, SystemLiteral.DTCS, front_wheel_speed, rear_wheel_speed)


def dtcs_drw(context: Context,
             front_wheel_speed: _OptionalSpeed,
             left_rear_wheel_speed: _OptionalSpeed,
             right_rear_wheel_speed: _OptionalSpeed) -> InterventionEvent:
    if front_wheel_speed < left_rear_wheel_speed:
        context.overwrite_throttle(0)
        return InterventionEvent(context, SystemLiteral.DTCS, "l", front_wheel_speed, left_rear_wheel_speed)
    if front_wheel_speed < right_rear_wheel_speed:
        context.overwrite_throttle(0)
        return InterventionEvent(context, SystemLiteral.DTCS, "r", front_wheel_speed, right_rear_wheel_speed)
    return InterventionExitEvent(context, SystemLiteral.DTCS,
                                 front_wheel_speed, left_rear_wheel_speed, right_rear_wheel_speed)


class DTCS(Plugin):
    def __init__(self) -> None:
        super(DTCS, self).__init__((["front_wheel_speed", "rear_wheel_speed"],
                                    ["front_wheel_speed", "left_rear_wheel_speed", "right_rear_wheel_speed"]))

    @_override
    def pre_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        context.intervene(dtcs_srw(context, kwargs["front_wheel_speed"], kwargs["rear_wheel_speed"])
                          if context.srw_mode() else
                          dtcs_drw(context, kwargs["front_wheel_speed"], kwargs["left_rear_wheel_speed"],
                                   kwargs["right_rear_wheel_speed"]))



from json import dumps as _dumps
from time import time as _time
from typing import Any as _Any

from leads import DataContainer as _DataContainer
from leads.comm import Server as _Server


class RuntimeData(object):
    start_time: int = int(_time())
    lap_time: list[int] = []
    comm: _Server | None = None

    def comm_notify(self, d: _DataContainer | dict[str, _Any]) -> None:
        if self.comm:
            self.comm.broadcast(d.encode() if isinstance(d, _DataContainer) else _dumps(d).encode())



from tkinter import Misc as _Misc, ARC as _ARC
from typing import override as _override, Literal as _Literal

from customtkinter import DoubleVar as _DoubleVar
from numpy import pi as _pi, sin as _sin, cos as _cos

from leads import require_config as _require_config
from leads_gui.prototype import parse_color, CanvasBased, TextBased, VariableControlled, autoscale_font
from leads_gui.types import Font as _Font, Color as _Color


class Speedometer(TextBased, VariableControlled):
    def __init__(self,
                 master: _Misc,
                 theme_key: str = "CTkButton",
                 width: float = 0,
                 height: float = 0,
                 variable: _DoubleVar | None = None,
                 style: _Literal[0, 1, 2] = 0,
                 next_style_on_click: bool = True,
                 maximum: float = 200,
                 font: tuple[_Font, _Font, _Font] | None = None,
                 text_color: _Color | None = None,
                 fg_color: _Color | None = None,
                 hover_color: _Color | None = None,
                 bg_color: _Color | None = None,
                 corner_radius: float | None = None) -> None:
        TextBased.__init__(self, master, theme_key, width, height, None, text_color, fg_color, hover_color, bg_color,
                           corner_radius, next_style_on_click,
                           lambda _: self.next_style() if next_style_on_click else lambda _: None)
        VariableControlled.__init__(self, variable if variable else _DoubleVar(master))
        self.attach(self.partially_render)
        self._style: _Literal[0, 1, 2] = style
        self._maximum: float = maximum
        cfg = _require_config()
        self._font: tuple[_Font, _Font, _Font] = font if font else (("Arial", cfg.font_size_x_large),
                                                                    ("Arial", cfg.font_size_large),
                                                                    ("Arial", cfg.font_size_small))
        self._font: tuple[_Font, _Font, _Font] = (
            autoscale_font(master, font[0]), autoscale_font(master, font[1]), autoscale_font(master, font[2])
        ) if font else (autoscale_font(master, ("Arial", cfg.font_size_x_large)),
                        autoscale_font(master, ("Arial", cfg.font_size_large)),
                        autoscale_font(master, ("Arial", cfg.font_size_small)))

    def next_style(self) -> None:
        self._style = (self._style + 1) % 3
        self.render()

    def style(self, style: _Literal[0, 1, 2] | None) -> _Literal[0, 1, 2] | None:
        if style is None:
            return self._style
        self._style = style
        self.render()

    @_override
    def dynamic_renderer(self, canvas: CanvasBased) -> None:
        canvas.clear("d")
        v = self._variable.get()
        w, h, hc, vc, limit = canvas.meta()
        font = self._font[self._style]
        if self._style > 0:
            r = min(hc, vc) * 1.25
            x, y = hc, vc + r * .25
            p = min(v / self._maximum, 1)
            rad = p * 4 * _pi / 3 - _pi / 6
            color = parse_color(("#" + str(hex(int(0xbf - p * 0xbf)))[2:] * 3,
                                 "#" + str(hex(int(0x4d + 0xb2 * p)))[2:] * 3))
            canvas.collect("d0", canvas.create_arc(x - r, y - r, x + r, y + r, start=-30, extent=240, width=4,
                                                   style=_ARC, outline=color))
            canvas.collect("d1", canvas.create_line(*(x, y) if self._style == 2 else (x - _cos(rad) * (r - 8),
                                                                                      y - _sin(rad) * (r - 8)),
                                                    x - _cos(rad) * (r + 8), y - _sin(rad) * (r + 8), width=4,
                                                    fill=color))
            canvas.collect("d2", canvas.create_text(x, y * .95 if self._style == 1 else y + (r - font[1]) * .5,
                                                    text=str(int(v)), fill=self._text_color, font=font))
        else:
            canvas.collect("d2", canvas.create_text(hc, vc, text=str(int(v)), fill=self._text_color, font=font))

    @_override
    def raw_renderer(self, canvas: CanvasBased) -> None:
        canvas.clear()
        w, h, hc, vc, limit = canvas.meta()
        font = self._font[self._style]
        if (target_font_size := h - 28 if self._style == 0 else h - 48) < font[1]:
            font = (font[0], target_font_size)
        canvas.draw_fg(self._fg_color, self._hover_color, self._corner_radius)
        if self._style > 0:
            r = min(hc, vc) * 1.25
            x, y = hc, vc + r * .25
            font_size_span = int(min(h * .06, self._font[2][1] * .8))
            canvas.collect("f1", canvas.create_text(x, y - r * .6, text="KM / H", fill="gray",
                                                    font=(font[0], font_size_span)))
            canvas.collect("f2", canvas.create_text(x - r * .7, y + r * .4, text="0", fill="gray",
                                                    font=(font[0], font_size_span)))
            canvas.collect("f3", canvas.create_text(x + r * .7, y + r * .4, text=str(self._maximum), fill="gray",
                                                    font=(font[0], font_size_span)))
        self.dynamic_renderer(canvas)



from typing import TypeVar as _TypeVar

from leads.context import Context
from leads.types import OnRegister as _OnRegister, OnRegisterChain as _OnRegisterChain

T = _TypeVar("T", bound=Context)

_context_instance: T | None = None

_on_register_context: _OnRegister[T] = lambda _: None


def set_on_register_context(callback: _OnRegisterChain[T]) -> None:
    global _on_register_context
    _on_register_context = callback(_on_register_context)


def register_context(context: T) -> None:
    global _context_instance
    if _context_instance:
        raise RuntimeError("Another context is already registered")
    _on_register_context(context)
    _context_instance = context


def get_context() -> T | None:
    return _context_instance


def require_context() -> T:
    if _context_instance:
        return _context_instance
    raise RuntimeError("No context registered")



from os.path import abspath as _abspath
from platform import system as _system


def get_system_platform() -> str:
    return _system().lower()


_ASSETS_PATH: str = _abspath(__file__)[:-9] + "assets"



from typing import Any as _Any

from leads import ConfigTemplate as _ConfigTemplate


class Config(_ConfigTemplate):
    def __init__(self, base: dict[str, _Any]) -> None:
        self.width: int = 720
        self.height: int = 480
        self.fullscreen: bool = False
        self.no_title_bar: bool = False
        self.manual_mode: bool = False
        self.refresh_rate: int = 30
        self.font_size_small: int = 14
        self.font_size_medium: int = 28
        self.font_size_large: int = 42
        self.font_size_x_large: int = 56
        self.comm_addr: str = "127.0.0.1"
        self.comm_port: int = 16900
        self.data_dir: str = "data"
        super().__init__(base)



from abc import abstractmethod as _abstractmethod, ABCMeta as _ABCMeta
from threading import Thread as _Thread
from typing import Any as _Any, override as _override, overload as _overload

from leads.os import _thread_flags


class Device(object):
    @_overload
    def __init__(self, *args, **kwargs) -> None:  # real signature unknown
        ...

    def __init__(self, *pins: int | str) -> None:
        self._tag: str = ""
        self._parent_tags: list[str] = []
        self._pins: tuple[int | str, ...] = pins

    def tag(self, tag: str | None = None) -> str | None:
        """
        Set or get the tag of the device.
        ```python
        tag(a_tag) # set a tag
        tag() # get the tag
        ```
        :param tag: the tag or None if getter mode
        :return: the tag or None if setter mode
        """
        if tag is None:
            return self._tag
        self._tag = tag

    def parent_tags(self, parent_tags: list[str] | None = None) -> list[str] | None:
        """
        Set or get the parent tags of the device.
        ```python
        parent_tags(a_parent_tag_list) # set the parent tags
        parent_tags() # get the parent tags
        ```
        :param parent_tags: the parent tags or None if getter mode
        :return: the parent tags or None if setter mode
        """
        if parent_tags is None:
            return self._parent_tags[:]
        if len(self._parent_tags) > 0:
            raise RuntimeError("Duplicated initialization")
        self._parent_tags = parent_tags

    def pins_check(self, required_num: int) -> None:
        if len(self._pins) != required_num:
            raise ValueError(f"`{self.__class__.__name__}` only takes in {required_num} pins")

    def initialize(self, *parent_tags: str) -> None:
        ...

    def read(self) -> _Any:
        raise NotImplementedError

    def write(self, payload: _Any) -> None:
        raise NotImplementedError

    def update(self, data: _Any) -> None:
        raise NotImplementedError

    def close(self) -> None:
        ...


class ShadowDevice(Device, metaclass=_ABCMeta):
    def __init__(self, *pins: int | str) -> None:
        super().__init__(*pins)
        self._shadow_thread: _Thread | None = None

    @_abstractmethod
    def loop(self) -> None:
        raise NotImplementedError

    def run(self) -> None:
        while _thread_flags.active:
            self.loop()

    @_override
    def initialize(self, *parent_tags: str) -> None:
        self._shadow_thread = _Thread(name=str(id(self)) + " shadow", target=self.run, daemon=True)
        self._shadow_thread.start()



from typing import override as _override

from pynmea2 import parse as _parse
from pynmea2.types.talker import TalkerSentence as _TalkerSentence
from serial import Serial as _Serial

from leads import Device as _Device, SFT as _SFT
from leads.comm import Entity as _Entity, Callback as _Callback, Service as _Service
from leads_comm_serial import SerialConnection as _SerialConnection


class NMEAGPSReceiver(_Device, _Entity):
    """
    Supports:
    - Any USB (serial) GPS receiver with NMEA 0183 output
    """
    def __init__(self, port: str, baud_rate: int = 9600) -> None:
        _Device.__init__(self, port)
        _Entity.__init__(self, -1, _NMEAGPSCallback(self))
        self._serial: _Serial = _Serial()
        self._serial.port = port
        self._serial.baudrate = baud_rate
        self._connection: _SerialConnection | None = None
        self._valid: bool = False
        self._ground_speed: float = 0
        self._latitude: float = 0
        self._longitude: float = 0
        self._quality: int = 0
        self._num_satellites: int = 0

    @_override
    def port(self) -> str:
        return self._serial.port

    @_override
    def initialize(self, *parent_tags: str) -> None:
        self.start(True)
        super().initialize(*parent_tags)

    @_override
    def update(self, data: _TalkerSentence) -> None:
        if hasattr(data, "latitude"):
            self._latitude = float(data.latitude)
        if hasattr(data, "longitude"):
            self._longitude = float(data.longitude)
        if hasattr(data, "is_valid"):
            if (v := data.is_valid) and not self._valid:
                _SFT.recover(self)
            elif not v and self._valid:
                _SFT.fail(self, "No fix")
            self._valid = v
        if NMEAGPSReceiver._has_field(data.fields, "spd_over_grnd", 6):
            self._ground_speed = float(ground_speed) * 1.852 if (ground_speed := data.data[6]) else 0
        if NMEAGPSReceiver._has_field(data.fields, "gps_qual", 5):
            self._quality = int(data.data[5])
        if NMEAGPSReceiver._has_field(data.fields, "num_sats", 6):
            self._num_satellites = int(data.data[6])

    @_override
    def read(self) -> [bool, float, float, float, int, int]:
        """
        :return: [validity, ground speed, latitude, longitude, quality, num of satellites]
        """
        return self._valid, self._ground_speed, self._latitude, self._longitude, self._quality, self._num_satellites

    @_override
    def run(self) -> None:
        self._callback.on_initialize(self)
        self._serial.open()
        self._callback.on_connect(self, connection := _SerialConnection(self, self._serial, self._serial.port,
                                                                        separator=b"\n"))
        self._connection = connection
        self._stage(connection)

    @_override
    def close(self) -> None:
        if self._connection:
            self._connection.close()

    @staticmethod
    def _has_field(fields: tuple[tuple[str, str, ...], ...], target_field: str, at: int) -> bool:
        return len(fields) > at and fields[at][1] == target_field


class _NMEAGPSCallback(_Callback):
    def __init__(self, receiver: NMEAGPSReceiver) -> None:
        super().__init__()
        self._receiver: _Device = receiver

    @_override
    def on_connect(self, service: _Service, connection: _SerialConnection) -> None:
        _SFT.fail(self._receiver, "No fix")

    @_override
    def on_receive(self, service: _Service, msg: bytes) -> None:
        self.super(service=service, msg=msg)
        self._receiver.update(_parse(msg.decode()))

    @_override
    def on_fail(self, service: _Service, error: Exception) -> None:
        self.super(service=service, error=error)
        _SFT.fail(self._receiver, error)



from abc import ABCMeta as _ABCMeta, abstractmethod as _abstractmethod
from collections import deque as _deque
from time import time as _time
from typing import TypeVar as _TypeVar, Generic as _Generic

from numpy import diff as _diff, average as _average, array as _array

from leads.constant import ESCMode
from leads.data import DataContainer, SRWDataContainer, DRWDataContainer

T = _TypeVar("T", bound=DataContainer)


def _check_data_type(data: T, superclass: type[DataContainer] = DataContainer) -> None:
    if not isinstance(data, superclass):
        raise TypeError(f"New data must inherit from `{superclass}`")


class Context(_Generic[T], metaclass=_ABCMeta):
    def __init__(self,
                 srw_mode: bool = True,
                 initial_data: T | None = None,
                 data_seq_size: int = 100,
                 num_laps_timed: int = 3) -> None:
        """
        :param srw_mode: True: single rear wheel mode; False: double rear wheel mode
        :param initial_data: initial data
        :param data_seq_size: buffer size of history data
        """
        self._srw_mode: bool = srw_mode
        dct = SRWDataContainer if srw_mode else DRWDataContainer
        if initial_data:
            _check_data_type(initial_data, dct)
        else:
            initial_data = dct()
        self._initial_data_type: type[DataContainer] = type(initial_data)
        if data_seq_size < 1:
            raise ValueError("`data_seq_size` must be greater or equal to 1")
        self._data_seq: _deque[dct] = _deque((initial_data,), maxlen=data_seq_size)
        self._speed_seq: _deque[float] = _deque(maxlen=data_seq_size)
        self._lap_time_seq: _deque[int] = _deque((int(_time() * 1000),), maxlen=num_laps_timed + 1)
        self._torque_mapping: list[float] = [1] if srw_mode else [1, 1]
        self._esc_mode: ESCMode = ESCMode.STANDARD
        self._hazard: bool = False

    def data(self) -> T:
        """
        :return: a copy of the current data container
        """
        return self._data_seq[-1]

    def push(self, data: T) -> None:
        """
        Push new data into the sequence.
        :param data: the new data
        """
        _check_data_type(data, self._initial_data_type)
        self._data_seq.append(data)
        self._speed_seq.append(data.speed)

    def srw_mode(self) -> bool:
        return self._srw_mode

    def esc_mode(self, esc_mode: ESCMode | None = None) -> ESCMode | None:
        if esc_mode is None:
            return self._esc_mode
        self._esc_mode = esc_mode

    @_abstractmethod
    def update(self) -> None:
        raise NotImplementedError

    @_abstractmethod
    def intervene(self, *args, **kwargs) -> None:  # real signature unknown
        raise NotImplementedError

    @_abstractmethod
    def suspend(self, *args, **kwargs) -> None:  # real signature unknown
        raise NotImplementedError

    def time_lap(self) -> None:
        self._lap_time_seq.append(int(_time() * 1000))

    def get_lap_time_list(self) -> list[int]:
        return [self._lap_time_seq[i] - self._lap_time_seq[i - 1] for i in range(1, len(self._lap_time_seq))]

    def get_speed_trend(self) -> float:
        return float(_average(_diff(_array(self._speed_seq)))) if len(self._speed_seq) > 1 else 0

    def torque_mapping(self, torque_mapping: list[float] | None = None) -> list[float] | None:
        if torque_mapping is None:
            return self._torque_mapping
        self._torque_mapping = torque_mapping

    def overwrite_throttle(self, force: float) -> float:
        # todo
        return 0

    def overwrite_brake(self, force: float) -> float:
        # todo
        return 0

    def hazard(self, hazard: bool | None = None) -> bool | None:
        if hazard is None:
            return self._hazard
        self._hazard = hazard



from leads.plugin.abs import *
from leads.plugin.atbs import *
from leads.plugin.dtcs import *
from leads.plugin.ebi import *
from leads.plugin.gps_speed_correction import *
from leads.plugin.plugin import *



from os import chmod as _chmod, getuid as _getuid
from os.path import abspath as _abspath
from pwd import getpwuid as _getpwuid


def create_service() -> None:
    _chmod(script := _abspath(__file__)[:-10] + "leads_vec.service.sh", 777)
    with open("/etc/systemd/system/leads_vec.service", "w") as f:
        f.write(
            "[Unit]\n"
            "Description=LEADS VeC\n"
            "After=display-manager.service\n"
            "Requires=display-manager.service\n"
            "[Service]\n"
            "Type=simple\n"
            f"User={(user := _getpwuid(_getuid()).pw_name)}\n"
            "Environment=\"DISPLAY=:0\"\n"
            f"Environment=\"XAUTHORITY=/home/{user}/.Xauthority\"\n"
            f"ExecStart=/bin/sh {script}\n"
            "[Install]\n"
            "WantedBy=graphical.target"
        )



from typing import override as _override

from leads.comm.prototype import Entity, Connection


class Client(Entity):
    """
    You should use `create_client()` and start_client()` instead of directly calling any method.
    """
    _connection: Connection | None = None

    @_override
    def run(self, server_address: str) -> None:
        self._callback.on_initialize(self)
        self._socket.connect((server_address, self._port))
        self._callback.on_connect(self, connection := Connection(self, self._socket, (server_address, self._port)))
        self._connection = connection
        self._stage(connection)

    def send(self, msg: bytes) -> None:
        if not self._connection:
            raise IOError("Client must be running to perform this operation")
        self._connection.send(msg)

    @_override
    def close(self) -> None:
        if self._connection:
            self._connection.close()



from importlib.util import find_spec as _find_spec

if not _find_spec("serial"):
    raise ImportError("Please install `pyserial` to run this module\n>>>pip install pyserial")

from leads_comm_serial.connection import *



from enum import StrEnum as _StrEnum, IntEnum as _IntEnum


class SystemLiteral(_StrEnum):
    DTCS: str = "DTCS"
    ABS: str = "ABS"
    EBI: str = "EBI"
    ATBS: str = "ATBS"


class ESCMode(_IntEnum):
    STANDARD: int = 0
    AGGRESSIVE: int = 1
    SPORT: int = 2
    OFF: int = 3



from threading import Thread as _Thread
from typing import override as _override

from leads.comm.prototype import Entity, Connection
from leads.os import _thread_flags


class Server(Entity):
    """
    You should use `create_server()` and start_server()` instead of directly calling any method.
    """
    _connections: list[Connection] = []

    def num_connections(self) -> int:
        return len(self._connections)

    def remove_connection(self, connection: Connection) -> None:
        try:
            self._connections.remove(connection)
        except ValueError:
            pass

    @_override
    def run(self, max_connection: int = 1) -> None:
        self._socket.bind(("0.0.0.0", self._port))
        self._socket.listen(max_connection)
        self._callback.on_initialize(self)
        while _thread_flags.active:
            socket, address = self._socket.accept()
            self._callback.on_connect(self, connection := Connection(self, socket, address,
                                                                     on_close=lambda c: self.remove_connection(c)))
            self._connections.append(connection)
            _Thread(target=self._stage, args=(connection,), daemon=True).start()

    def broadcast(self, msg: bytes) -> None:
        for c in self._connections:
            try:
                c.send(msg)
            except IOError:
                self.remove_connection(c)

    @_override
    def close(self) -> None:
        self._socket.close()
        for connection in self._connections:
            connection.close()



from argparse import ArgumentParser as _ArgumentParser, BooleanOptionalAction as _BooleanOptionalAction
from os import mkdir as _mkdir, chmod as _chmod
from os.path import exists as _exists
from subprocess import run as _run
from sys import exit as _exit, version as _version

from leads import register_controller as _register_controller, MAIN_CONTROLLER as _MAIN_CONTROLLER, \
    L as _L, load_config as _load_config, register_config as _register_config, device as _device, \
    GPS_RECEIVER as _GPS_RECEIVER, reset as _reset
from leads_gui import Config as _Config
from leads_gui.system import get_system_platform as _get_system_platform

if __name__ == "__main__":
    parser = _ArgumentParser(prog="LEADS VeC",
                             description="Lightweight Embedded Assisted Driving System VeC",
                             epilog="ProjectNeura: https://projectneura.org"
                                    "GitHub: https://github.com/ProjectNeura/LEADS")
    parser.add_argument("action", choices=("info", "run"))
    parser.add_argument("-r", "--register", choices=("systemd", "config"), default=None, help="service to register")
    parser.add_argument("-c", "--config", default=None, help="specified configuration file")
    parser.add_argument("--emu", action=_BooleanOptionalAction, default=False, help="use emulator")
    parser.add_argument("--xws", action=_BooleanOptionalAction, default=False, help="use X Window System")
    args = parser.parse_args()
    if args.action == "info":
        from leads_vec.__version__ import __version__

        _L.info("LEADS Version: " + __version__,
                "System Platform: " + _get_system_platform(),
                "Python Version: " + _version,
                sep="\n")
        _exit()
    if args.register == "systemd":
        if _get_system_platform() != "linux":
            _exit("Error: Unsupported operating system")
        if not _exists("/usr/local/leads/config.json"):
            _L.info("Config file not found. Creating \"/usr/local/leads/config.json\"...")
            _mkdir("/usr/local/leads")
            with open("/usr/local/leads/config.json", "w") as f:
                f.write(str(_Config({})))
            _L.info("Using \"/usr/local/leads/config.json\"")
        _chmod("/usr/local/leads/config.json", 777)
        from ._bootloader import create_service

        create_service()
        _L.info("Service registered")
    elif args.register == "config":
        if _exists("config.json"):
            r = input("\"config.json\" already exists. Overwrite? (y/N) >>>").lower()
            if r.lower() != "y":
                _exit("Error: Aborted")
        with open("config.json", "w") as f:
            f.write(str(_Config({})))
        _L.info("Configuration file saved to \"config.json\"")
    _register_config(config := _load_config(args.config, _Config) if args.config else _Config({}))
    _L.debug("Configuration loaded:", str(config))
    if args.xws:
        from os import getuid as _getuid
        from pwd import getpwuid as _getpwuid

        _L.info("Configuring X Window System...")
        _run(("/usr/bin/xhost", "+SI:localuser:" + _getpwuid(_getuid()).pw_name))
    try:
        if args.emu:
            raise AttributeError("User specifies to use emulator")
        from leads_vec.controller import _
    except (ImportError, AttributeError) as e:
        _reset()
        _L.debug(repr(e))
        if isinstance(e, ImportError):
            _L.warn("`leads_vec.controller` is not available, using emulation module instead...")
        try:
            if config.srw_mode:
                from leads_emulation import SRWSin as _Controller
            else:
                from leads_emulation import DRWSin as _Controller

            _register_controller(_MAIN_CONTROLLER, _Controller())
            from leads_emulation import GPSReceiver as _GPSReceiver


            @_device(_GPS_RECEIVER, _MAIN_CONTROLLER)
            class GPS(_GPSReceiver):
                pass
        except ImportError:
            raise ImportError("At least one adapter has to be installed")
    from leads_vec.cli import main

    _exit(main())



from copy import copy as _copy
from operator import add as _add, sub as _sub, mul as _mul, truediv as _truediv, floordiv as _floordiv, lt as _lt, \
    le as _le, gt as _gt, ge as _ge
from typing import TextIO as _TextIO, TypeVar as _TypeVar, Generic as _Generic, Sequence as _Sequence, \
    override as _override, Self as _Self, Iterator as _Iterator, Callable as _Callable

from numpy import mean as _mean, array as _array
from numpy.linalg import norm as _norm

from leads.types import Compressor as _Compressor, Stringifier as _Stringifier

T = _TypeVar("T")


def mean_compressor(sequence: list[T], target_size: int) -> list[T]:
    """
    A compression method that reduces data memory usage by averaging adjacent numbers and merging them.
    :param sequence: the sequence to compress
    :param target_size: expected size
    :return: the compressed sequence
    """
    chunk_size = int(len(sequence) / target_size)
    if chunk_size < 2:
        return sequence
    r = []
    for i in range(target_size - 1):
        r.append(_mean(sequence[i * chunk_size: (i + 1) * chunk_size]))
    r.append(_mean(sequence[(target_size - 1) * chunk_size:]))
    return r


def csv_stringifier(element: T) -> str:
    """
    Dump an element as a CSV string.
    :param element: the element to stringify
    :return: CSV string
    """
    return str(element) + ","


class DataPersistence(_Sequence[T], _Generic[T]):
    def __init__(self,
                 file: str | _TextIO | None = None,
                 max_size: int = -1,
                 chunk_scale: int = 1,
                 persistence: bool = False,
                 compressor: _Compressor[T] = mean_compressor,
                 stringifier: _Stringifier[T] = csv_stringifier) -> None:
        """
        :param file: the file into which the data is written
        :param max_size: maximum cached size
        :param chunk_scale: chunk scaling factor (compression)
        :param compressor: compressor interface
        :param stringifier: stringifier interface
        """
        self._file: _TextIO | None = open(file, "a") if isinstance(file, str) else file
        self._max_size: int = max_size
        self._chunk_scale: int = chunk_scale
        self._persistence: bool = persistence if file else False
        self._compressor: _Compressor[T] = compressor
        self._stringifier: _Stringifier[T] = stringifier
        self._data: list[T] = []
        self._chunk: list[T] = []
        self._chunk_size: int = chunk_scale

    @_override
    def __len__(self) -> int:
        return len(self._data)

    @_override
    def __getitem__(self, item: int | slice) -> T | list[T]:
        return self._data[item]

    @_override
    def __str__(self) -> str:
        return str(self._data)

    def close(self) -> None:
        if self._file:
            self._file.close()

    def get_chunk_size(self) -> int:
        return self._chunk_size

    def to_list(self) -> list[T]:
        return _copy(self._data)

    def get_chunk(self) -> list[T]:
        return _copy(self._chunk)

    def _push_to_data(self, element: T) -> None:
        self._data.append(element)
        if self._max_size < 2:
            return
        if len(self._data) >= self._max_size:
            self._data = self._compressor(self._data, int(len(self._data) * .5))
            self._chunk_size *= 2

    def append(self, element: T) -> None:
        if self._persistence:
            self._file.write(self._stringifier(element))
        if self._chunk_size == 1:
            return self._push_to_data(element)
        self._chunk.append(element)
        if len(self._chunk) >= self._chunk_size:
            for e in self._compressor(self._chunk, self._chunk_scale):
                self._push_to_data(e)
            self._chunk.clear()


E = _TypeVar("E")


class Vector(_Sequence[E], _Generic[E]):
    def __init__(self, *coordinates: E) -> None:
        self._d: int = len(coordinates)
        self._coordinates: tuple[E, ...] = coordinates

    @_override
    def __len__(self) -> int:
        return self._d

    @_override
    def __iter__(self) -> _Iterator[E]:
        return iter(self._coordinates)

    @_override
    def __getitem__(self, item: int | slice) -> _Self:
        return Vector(*self._coordinates[item])

    @_override
    def __str__(self) -> str:
        return ";".join(map(str, self._coordinates))

    @_override
    def __eq__(self, other: _Self) -> bool:
        return self._coordinates == other._coordinates

    def _check_dimension(self, other: _Self) -> None:
        if other._d != self._d:
            raise ValueError("Cannot perform this operation on two vectors of different dimensions")

    def __neg__(self) -> _Self:
        return Vector(*(-i for i in self._coordinates))

    def __abs__(self) -> _Self:
        return Vector(*(abs(i) for i in self._coordinates))

    def _operate(self, other: _Self | E, operator: _Callable[[E, E], E]) -> _Self:
        if isinstance(other, Vector):
            self._check_dimension(other)
            return Vector(*(operator(self._coordinates[i], other._coordinates[i]) for i in range(self._d)))
        return Vector(*(operator(i, other) for i in self._coordinates))

    def __add__(self, other: _Self | E) -> _Self:
        return self._operate(other, _add)

    def __sub__(self, other: _Self | E) -> _Self:
        return self._operate(other, _sub)

    def __mul__(self, other: _Self | E) -> _Self:
        return self._operate(other, _mul)

    def __truediv__(self, other: _Self | E) -> _Self:
        return self._operate(other, _truediv)

    def __floordiv__(self, other: _Self | E) -> _Self:
        return self._operate(other, _floordiv)

    def distance(self, other: _Self) -> float:
        return float(_norm(_array(self._coordinates) - _array(other._coordinates)))

    def magnitude(self) -> float:
        return self.distance(Vector(*(0,) * self._d))

    def _compare(self, other: _Self | E, comparer: _Callable[[E, E], bool]) -> bool:
        return comparer(self.magnitude(), other.magnitude() if isinstance(other, Vector) else other)

    def __lt__(self, other: _Self | E) -> bool:
        return self._compare(other, _lt)

    def __le__(self, other: _Self | E) -> bool:
        return self._compare(other, _le)

    def __gt__(self, other: _Self | E) -> bool:
        return self._compare(other, _gt)

    def __ge__(self, other: _Self | E) -> bool:
        return self._compare(other, _ge)



import sys as _sys
from threading import Lock as _Lock
from types import FrameType as _FrameType

if hasattr(_sys, "_getframe"):
    def _currentframe() -> _FrameType:
        return getattr(_sys, "_getframe")(1)
else:
    def _currentframe() -> _FrameType:
        try:
            raise Exception
        except Exception as exc:
            return exc.__traceback__.tb_frame.f_back


class _ThreadFlags(object):
    def __init__(self) -> None:
        self._lock: _Lock = _Lock()
        self._active: bool = True

    @property
    def active(self) -> bool:
        return self._active

    @active.setter
    def active(self, active: bool) -> None:
        self._lock.acquire()
        try:
            self._active = active
        finally:
            self._lock.release()


_thread_flags: _ThreadFlags = _ThreadFlags()



from abc import ABCMeta as _ABCMeta
from json import dumps as _dumps
from time import time as _time
from typing import override as _override, Any as _Any


class Serializable(object):
    def to_dict(self) -> dict[str, _Any]:
        return {n: v for n in dir(self) if not n.startswith("_") and not callable(v := getattr(self, n))}


class DataContainer(Serializable, metaclass=_ABCMeta):
    def __init__(self,
                 voltage: float,
                 min_speed: float,
                 forward_acceleration: float,
                 lateral_acceleration: float,
                 mileage: float,
                 gps_valid: bool,
                 gps_ground_speed: float,
                 latitude: float,
                 longitude: float) -> None:
        self._time_stamp: int = int(_time() * 1000)
        self.voltage: float = voltage
        self.speed: float = min_speed
        self.forward_acceleration: float = forward_acceleration
        self.lateral_acceleration: float = lateral_acceleration
        self.mileage: float = mileage
        self.gps_valid: bool = gps_valid
        self.gps_ground_speed: float = gps_ground_speed
        self.latitude: float = latitude
        self.longitude: float = longitude

    @_override
    def __str__(self) -> str:
        return _dumps(self.to_dict())

    def reset_time_stamp(self) -> None:
        """
        Reset the data's time stamp to now.
        """
        self._time_stamp = int(_time() * 1000)

    def get_time_stamp(self) -> int:
        """
        Get the data's time stamp.
        :return: the time stamp (ms)
        """
        return self._time_stamp

    @_override
    def to_dict(self) -> dict:
        """
        Convert the data into a dictionary.
        :return: a dictionary that contains all custom attributes of the container
        """
        (r := super().to_dict())["t"] = self._time_stamp
        return r

    def encode(self) -> bytes:
        """
        Encode the data into bytes for network transaction purposes.
        :return: JSON in bytes
        """
        return str(self).encode()


class SRWDataContainer(DataContainer):
    def __init__(self,
                 voltage: float = 0,
                 min_speed: float = 0,
                 forward_acceleration: float = 0,
                 lateral_acceleration: float = 0,
                 mileage: float = 0,
                 gps_valid: bool = False,
                 gps_ground_speed: float = 0,
                 latitude: float = 0,
                 longitude: float = 0,
                 front_wheel_speed: float = 0,
                 rear_wheel_speed: float = 0) -> None:
        super().__init__(voltage, min_speed, forward_acceleration, lateral_acceleration, mileage, gps_valid,
                         gps_ground_speed, latitude, longitude)
        self.front_wheel_speed: float = front_wheel_speed
        self.rear_wheel_speed: float = rear_wheel_speed


class DRWDataContainer(DataContainer):
    def __init__(self,
                 voltage: float = 0,
                 min_speed: float = 0,
                 forward_acceleration: float = 0,
                 lateral_acceleration: float = 0,
                 mileage: float = 0,
                 gps_valid: bool = False,
                 gps_ground_speed: float = 0,
                 latitude: float = 0,
                 longitude: float = 0,
                 front_wheel_speed: float = 0,
                 left_rear_wheel_speed: float = 0,
                 right_rear_wheel_speed: float = 0) -> None:
        super().__init__(voltage, min_speed, forward_acceleration, lateral_acceleration, mileage, gps_valid,
                         gps_ground_speed, latitude, longitude)
        self.front_wheel_speed: float = front_wheel_speed
        self.left_rear_wheel_speed: float = left_rear_wheel_speed
        self.right_rear_wheel_speed: float = right_rear_wheel_speed



from leads.comm.client.client import Client
from leads.comm.prototype import Callback


def create_client(port: int = 16900, callback: Callback = Callback()) -> Client:
    return Client(port, callback)


def start_client(server_address: str, target: Client = create_client(), parallel: bool = False) -> Client:
    return target.start(parallel, server_address=server_address)



from json import load as _load
from typing import TypeVar as _TypeVar, TextIO as _TextIO

from leads.config.template import ConfigTemplate
from leads.types import OnRegister as _OnRegister, OnRegisterChain as _OnRegisterChain

T = _TypeVar("T", bound=ConfigTemplate)

_config_instance: T | None = None

_on_register_config: _OnRegister[T] = lambda _: None


def set_on_register_config(callback: _OnRegisterChain[T]) -> None:
    global _on_register_config
    _on_register_config = callback(_on_register_config)


def load_config(file: str | _TextIO, constructor: type[T]) -> T:
    if isinstance(file, str):
        with open(file) as f:
            return constructor(_load(f))
    return constructor(_load(file))


def register_config(config: T) -> None:
    global _config_instance
    if _config_instance:
        raise RuntimeError("Another config is already registered")
    _on_register_config(config)
    _config_instance = config


def get_config() -> T | None:
    return _config_instance


def require_config() -> T:
    if _config_instance:
        return _config_instance
    raise RuntimeError("No config registered")



from time import time as _time

from leads import require_config as _require_config


class PerformanceChecker(object):
    def __init__(self) -> None:
        self._refresh_rate: int = _require_config().refresh_rate
        self._fps: float = 0
        self._delay_offset: float = 0
        self._last_frame: float = _time()

    def fps(self) -> float:
        return self._fps

    def record_frame(self) -> None:
        # add .001 to avoid zero division
        delay = .001 + (t := _time()) - self._last_frame
        self._fps = 1 / delay
        self._delay_offset = (self._delay_offset * (self._refresh_rate - 1) + delay) / self._refresh_rate
        self._last_frame = t

    def delay_offset(self) -> float:
        return self._delay_offset * 1000



from typing import override as _override, overload as _overload

from leads.dt.device import Device


class Controller(Device):
    @_overload
    def __init__(self, *args, **kwargs) -> None:  # real signature unknown
        ...

    def __init__(self) -> None:
        super().__init__()
        self._devices: dict[str, Device] = {}

    def level(self) -> int:
        return len(self._parent_tags)

    def _attach_device(self, tag: str, device: Device) -> None:
        self._devices[tag] = device
        device.tag(tag)

    def devices(self) -> list[Device]:
        """
        :return: the device list
        """
        return list(self._devices.values())

    def device(self, tag: str, device: Device | None = None) -> Device | None:
        """
        Set or get a device by tag. The device's tag will be overwritten.
        ```python
        device(a_tag, a_device) # set a device
        device(a_tag) # get the device
        ```
        :param tag: tag of the device (it shares the global namespace)
        :param device: the device or None if getter mode
        :return: the device or None if setter mode
        """
        if device is None:
            return self._devices[tag]
        self._attach_device(tag, device)

    @_override
    def initialize(self, *parent_tags: str) -> None:
        for device in self._devices.values():
            device.initialize(*self._parent_tags, self._tag)



from atexit import register as _register

from leads.config import *
from leads.context import *
from leads.data import *
from leads.dt import *
from leads.event import *
from leads.leads import *
from leads.logger import Level, L
from leads.os import _thread_flags
from leads.plugin import *
from leads.registry import *
from leads.sft import SFT, mark_system, read_marked_system


@_register
def _request_threads_stop() -> None:
    _thread_flags.active = False



from setuptools import find_packages, setup

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="leads",
    version="0.8.11",
    python_requires=">=3.12",
    author="ProjectNeura",
    author_email="central@projectneura.org",
    description="Lightweight Embedded Assisted Driving System",
    license="Apache License 2.0",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/ProjectNeura/LEADS",
    packages=find_packages(),
    package_data={
        "leads_gui": ["assets/*", "assets/icons/*"],
        "leads_vec": ["_bootloader/leads_vec.service.sh"]
    },
    include_package_data=True,
    install_requires=["numpy"]
)



from typing import Any as _Any

from leads.constant import ESCMode
from leads.context import Context
from leads.registry import require_context
from leads.types import RequiredData as _RequiredData


class Plugin(object):
    def __init__(self, required_data: _RequiredData) -> None:
        """
        :param required_data: required data or (required data in srw mode, required data in drw mode)
        """
        super().__init__()
        self._required_data: _RequiredData = required_data
        self.state: dict[str, _Any] = {}
        self._enabled: bool = True

    def __getitem__(self, key: str) -> _Any:
        return self.state[key]

    def __setitem__(self, key: str, value: _Any) -> None:
        self.state[key] = value

    def enabled(self, enabled: bool | None = None) -> bool | None:
        if enabled is None:
            return require_context().esc_mode() != ESCMode.OFF and self._enabled
        self._enabled = enabled

    def required_data(self) -> list[str]:
        return (self._required_data if isinstance(self._required_data, list) else
                self._required_data[0] if require_context().srw_mode() else self._required_data[1])

    def on_load(self, context: Context) -> None: ...

    def pre_push(self, context: Context, kwargs: dict[str, _Any]) -> None:
        """
        Note that the new data at this point is not available yet.
        :param context: target context
        :param kwargs: {required data: its value}
        """
        ...

    def post_push(self, context: Context, kwargs: dict[str, _Any]) -> None:
        """
        :param context: target context
        :param kwargs: {required data: its value}
        """
        ...

    def pre_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        """
        :param context: target context
        :param kwargs: {required data: its value}
        """
        ...

    def post_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        """
        :param context: target context
        :param kwargs: {required data: its value}
        """
        ...



from typing import override as _override

from serial import Serial as _Serial

from leads import Controller as _Controller, SFT as _SFT
from leads.comm import Entity as _Entity, Callback as _Callback, Service as _Service
from leads_comm_serial import SerialConnection as _SerialConnection


class ArduinoProto(_Controller, _Entity):
    """
    Supports:
    - Any arduino connected through a USB (serial) port
    """

    def __init__(self, port: str, baud_rate: int = 9600) -> None:
        _Controller.__init__(self)
        _Entity.__init__(self, -1, _ArduinoCallback(self))
        self._serial: _Serial = _Serial()
        self._serial.port = port
        self._serial.baudrate = baud_rate
        self._connection: _SerialConnection | None = None

    @_override
    def port(self) -> str:
        return self._serial.port

    @_override
    def initialize(self, *parent_tags: str) -> None:
        self.start(True)
        super().initialize(*parent_tags)

    @_override
    def update(self, data: str) -> None:
        for d in self.devices():
            d.update(data)

    @_override
    def run(self) -> None:
        self._callback.on_initialize(self)
        self._serial.open()
        self._callback.on_connect(self, connection := _SerialConnection(self, self._serial, self._serial.port))
        self._connection = connection
        self._stage(connection)

    @_override
    def write(self, payload: bytes) -> None:
        if not self._connection:
            raise IOError("Target must be connected to perform this operation")
        self._connection.send(payload)

    @_override
    def close(self) -> None:
        if self._connection:
            self._connection.close()


class _ArduinoCallback(_Callback):
    def __init__(self, arduino: ArduinoProto) -> None:
        super().__init__()
        self._arduino: ArduinoProto = arduino

    @_override
    def on_receive(self, service: _Service, msg: bytes) -> None:
        self.super(service=service, msg=msg)
        self._arduino.update(msg.decode())

    @_override
    def on_fail(self, service: _Service, error: Exception) -> None:
        self.super(service=service, error=error)
        _SFT.fail(self._arduino, error)



from typing import Any as _Any, override as _override

from leads.callback import CallbackChain
from leads.context import Context
from leads.data import DataContainer


class Event(object):
    def __init__(self, t: str, context: Context) -> None:
        self.t: str = t
        self.context: Context = context

    @_override
    def __str__(self) -> str:
        return f"[{self.t}] {self.context.data()}"


class DataPushedEvent(Event):
    def __init__(self, context: Context, data: DataContainer) -> None:
        super().__init__("DATA PUSHED", context)
        self.data: DataContainer = data


class UpdateEvent(Event):
    def __init__(self, context: Context) -> None:
        super().__init__("UPDATE", context)


class SystemEvent(Event):
    def __init__(self, t: str, context: Context, system: str) -> None:
        super().__init__(t, context)
        self.system: str = system


class InterventionEvent(SystemEvent):
    def __init__(self, context: Context, system: str, *data: _Any) -> None:
        super().__init__("INTERVENTION", context, system)
        self.data: tuple[_Any, ...] = data


class InterventionExitEvent(InterventionEvent):
    pass


class SuspensionEvent(SystemEvent):
    def __init__(self, context: Context, system: str, cause: str, fatal: bool = False) -> None:
        super().__init__("SUSPENSION", context, system)
        self.cause: str = cause
        self.fatal: bool = fatal


class EventListener(CallbackChain):
    @_override
    def super(self, e: Event) -> None:
        super().super(e)

    def pre_push(self, event: DataPushedEvent) -> None: ...

    def post_push(self, event: DataPushedEvent) -> None: ...

    def on_update(self, event: UpdateEvent) -> None: ...

    def pre_intervene(self, event: InterventionEvent) -> None: ...

    def post_intervene(self, event: InterventionExitEvent) -> None: ...

    def pre_suspend(self, event: SuspensionEvent) -> None: ...

    def post_suspend(self, event: SuspensionEvent) -> None: ...



from typing import override as _override, Any as _Any

from leads.context import Context
from leads.plugin.plugin import Plugin


class GPSSpeedCorrection(Plugin):
    def __init__(self) -> None:
        super().__init__(["speed", "gps_valid", "gps_ground_speed"])
        self._loss: float = 0
        self._abs_loss: float = 0
        self._n: int = 0

    @_override
    def pre_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        d = context.data()
        if not d.gps_valid:
            return
        self._loss += (current_loss := d.speed - d.gps_ground_speed)
        self._abs_loss += abs(current_loss)
        self._n += 1
        if d.speed < 5 and current_loss < self._abs_loss / self._n:
            d.speed = d.gps_ground_speed + self._loss / self._n



MAIN_CONTROLLER: str = "__main__"

WHEEL_SPEED_CONTROLLER: str = "wsc"
LEFT_FRONT_WHEEL_SPEED_SENSOR: str = "wss_lf"
RIGHT_FRONT_WHEEL_SPEED_SENSOR: str = "wss_rf"
CENTER_REAR_WHEEL_SPEED_SENSOR: str = "wss_cr"
LEFT_REAR_WHEEL_SPEED_SENSOR: str = "wss_lr"
RIGHT_REAR_WHEEL_SPEED_SENSOR: str = "wss_rr"
ODOMETER: str = "odm"
POWER_CONTROLLER: str = "pc"
THROTTLE_PEDAL: str = "tpd"
BRAKE_PEDAL: str = "bpd"
DC_MOTOR_CONTROLLER_A: str = "dmca"
DC_MOTOR_CONTROLLER_B: str = "dmcb"
GPS_RECEIVER: str = "gps"
LEFT_INDICATOR: str = "lindi"
RIGHT_INDICATOR: str = "rindi"
